import numpy as np
import pandas as pd
import torch
import itertools
import random
import os
import json
import pathlib
from vae_gumbel_softmax3 import VAE_gumbel_hp_tuning, train, test, latent_repr, evaluate
from classification_latent import classification_latent_space
import argparse
import matplotlib.pyplot as plt
from load_data import load_dataset
from numpy import loadtxt, savetxt


parser = argparse.ArgumentParser(description='PyTorch VAE Gumbel Softmax')
parser.add_argument('-nc', '--no_cuda', action='store_true', default=False,
                    help='disables CUDA training')
parser.add_argument('--manual_seed', type=int, help='manual seed, if not given resorts to random seed.')
fp = parser.add_mutually_exclusive_group(required=False)
parser.add_argument('-li', '--log_interval', type=int, default=10, metavar='LOG_INTERVAL',
                    help='how many batches to wait before logging training status')

fp.add_argument('-te', '--testing', action='store_true', dest='testing',
                help='evaluate on test set after training')
fp.add_argument('-va', '--validation', action='store_false', dest='testing',
                help='only evaluate on validation set')
parser.set_defaults(testing=False)
parser.add_argument('-bs', '--batch_size', type=int, default=128, metavar='BATCH_SIZE',
                    help='input batch size for training (default: 100)')
parser.add_argument('--length', type=int, default=80, metavar='FSSIZE',
                    help='size of flattened sequence')
parser.add_argument('--encoder_dim', type=int, default=40, metavar='ESIZE',
                    help='output feature dim of encoder')
parser.add_argument('--decoder_dim', type=int, default=480, metavar='DSIZE',
                    help='output feature dim of decoder') 

parser.add_argument('--cnn_enc', type=bool, default=True, metavar='cnn_enc',
                    help='encoder cnn')
parser.add_argument('--test_model', type=bool, default=False, metavar='test_model',
                    help='test model')
parser.add_argument('--data', type=str, default='before', metavar='data',
                    help='training data type')
parser.add_argument('--only_positives', type=bool, default=False, metavar='test_model',
                    help='train VAE only on positives')
parser.add_argument('--only_negatives', type=bool, default=False, metavar='test_model',
                    help='train VAE only on positives')
# gpu/cpu
parser.add_argument('--gpu_num', type=int, default=0, metavar='GPU', help='choose GPU to run on.')

parser.add_argument('--incl_rec', type=bool, default=False, metavar='incl_rec',
                    help='include reconstruction error')
parser.add_argument('--act', type=str, default='sigmoid', metavar='act',
                    help='activation of decoder')
args = parser.parse_args()
args.cuda = not args.no_cuda and torch.cuda.is_available()
args.manual_seed = 21018
if args.manual_seed is None:
    args.manual_seed = random.randint(1, 100000)
random.seed(args.manual_seed)
torch.manual_seed(args.manual_seed)
np.random.seed(args.manual_seed)

print("seed", args.manual_seed)

if args.cuda:
    torch.cuda.set_device(args.gpu_num)

    
curr_dir_path = str(pathlib.Path().absolute())
save_dir_path = '/cnn' + '/logs/'
data_dir_path = '/data/'
cwd = os.getcwd() 
FILE_NAME = 'VAE_hyp_tuning.csv'
print(args.act)

args.gpu_num = 1

if __name__ == '__main__':

#    with open(curr_dir_path + "/config.json", encoding='utf-8', errors='ignore') as json_data:
#        config = json.load(json_data, strict=False)


    hyp_batch_size =  [256]#[256]
    hyp_no_filters1 = [16]# [16, 25] #[8, 12, 16, 25]# [8, 16, 25]# [8, 16, 25]#, 16, 32] #[8, 12, 16] #[16, 32, 64]
    hyp_no_filters2 =  [16] #, 16, 32] #[4, 8, 10] # 12]
    hyp_filter_size = [(1,7)]# [(1,7), (1,5), (1,9)]#[(1,5), (1,7)]#, (1, 5), (1,7)]#[3, 4, 5]#, 5, 7]
    hyp_filter2_size =  [(1,5)]#[3, 5, 7]#hyp_filter_size # [5] 
    hyp_stride =   [(1,1)]#[(1,2), (1,1)]#[1]#, 2]
    hyp_pool_size =  [(4,3)]# [(4,3), (4,2), (4,1)]#[2, 3]#[2, 3]#, 3]
    hyp_pool_stride = [(1,3)]# [1, 2]
    hyp_lr = [1e-5] #[1e-5, 1e-6]# [1e-6, 1e-5]# [1e-3, 1e-4, 1e-5]
    hyp_latent_dim =[ 15, 20, 30]# [2, 3, 4, 5, 6]#, 4, 5, 6, 7, 8]
    hyp_categorical_dim = [2]#[2, 3]#, 3, 4]
    hyp_temp_anneal =  [0.000005]# [0.3, 0.5, 0.7]#[0.0000005]#[0.0000001, 0.0000005]#[0.00000005, 0.0000001, 0.0000005] #[1.5, 1.0, 0.7]#, 2, 3] #[0.7, 1.0, 1.5]
    hym_num_conv =  [0.0, 0.2, 0.5]#[1, 2]
    
    list_hyperparam = [hyp_batch_size, hyp_no_filters1, hyp_filter_size, hyp_stride, hyp_pool_size, hyp_pool_stride, hyp_lr, hyp_latent_dim, hyp_categorical_dim, hyp_temp_anneal, hym_num_conv, hyp_filter2_size, hyp_no_filters2]
    

    
    """   
    hyp_batch_size =  [256]#[256]
    hyp_no_filters1 = [16, 32, 64] #[8, 16] #[8, 12, 16, 25]# [8, 16, 25]# [8, 16, 25]#, 16, 32] #[8, 12, 16] #[16, 32, 64]
    hyp_no_filters2 =  [16] #, 16, 32] #[4, 8, 10] # 12]
    hyp_filter_size =  [(1,7),(1,5)] #[(1,5), (1,7)]#[(1,5), (1,7)]#, (1, 5), (1,7)]#[3, 4, 5]#, 5, 7]
    hyp_filter2_size =  [(1,7)]#[3, 5, 7]#hyp_filter_size # [5] 
    hyp_stride =   [(1,2),(1,1), (1,1)] #[(1,2), (1,3)]#[(1,2), (1,1)]#[1]#, 2]
    hyp_pool_size =  [(1,2), (1,3)] #[2, 3]#[2, 3]#[2, 3]#, 3]
    hyp_pool_stride = [(2,2)]# [1, 2]
    hyp_lr = [1e-5] #[1e-5, 1e-6]# [1e-6, 1e-5]# [1e-3, 1e-4, 1e-5]
    hyp_latent_dim =[4, 5, 6] # 4, 5, 6]# [2, 3, 4, 5, 6]#, 4, 5, 6, 7, 8]
    hyp_categorical_dim = [2, 3, 4] #[2, 3]#[2, 3]#, 3, 4]
    hyp_temp_anneal =  [0.0000005]#[0.0000001, 0.0000005]#[0.00000005, 0.0000001, 0.0000005] #[1.5, 1.0, 0.7]#, 2, 3] #[0.7, 1.0, 1.5]
    hym_num_conv =  [1, 2]#[2]
    """
    ###pool size-pool stride always equal / num filters1-num filters2 always equal

    list_hyperparam = [hyp_batch_size, hyp_no_filters1, hyp_filter_size, hyp_stride, hyp_pool_size, hyp_pool_stride, hyp_lr, hyp_latent_dim, hyp_categorical_dim, hyp_temp_anneal, hym_num_conv, hyp_filter2_size, hyp_no_filters2]
    
    print("list of hp", list_hyperparam)
    
    grid = list(itertools.product(*list_hyperparam))
    models_no = len(grid)
    print(f"The total number of models is {models_no}")
    random.shuffle(grid)
    hyp_file_write_path = cwd + save_dir_path + FILE_NAME 
    
    #### Load data for the VAE training ####
    if args.data == 'before':
        print("training data type: sequences before the repeat")
        args.length = 40
        
        neg_seq_encoded_total = loadtxt(cwd + data_dir_path + 'before_data/' + 'chr19.test_negative_before.csv', delimiter=',')
        neg_seq_encoded_total2 = loadtxt(cwd +  data_dir_path + 'before_data/' + 'chr19.test_negative_before_2.csv', delimiter=',')
        pos_seq_encoded_chr19 = loadtxt(cwd +  data_dir_path + 'before_data/' + 'chr19.test_positive_before.csv', delimiter=',')
        pos_seq_encoded_chr21 = loadtxt(cwd +  data_dir_path + 'before_data/' + 'chr21.test_positive_before.csv', delimiter=',')
        pos_seq_encoded_chr22 = loadtxt(cwd +  data_dir_path + 'before_data/' + 'chr22.test_positive_before.csv', delimiter=',')
        pos_seq_encoded_chr7 = loadtxt(cwd +  data_dir_path + 'before_data/' + 'chr7.test_positive_before.csv', delimiter=',')
        pos_seq_encoded_chr8 = loadtxt(cwd +  data_dir_path + 'before_data/' + 'chr8.test_positive_before.csv', delimiter=',')
        pos_seq_encoded_total = np.concatenate((pos_seq_encoded_chr19, pos_seq_encoded_chr21, pos_seq_encoded_chr22, pos_seq_encoded_chr7, pos_seq_encoded_chr8), axis = 0)
        neg_seq_encoded_total = np.append(neg_seq_encoded_total, neg_seq_encoded_total2, axis = 0)
        neg_seq_encoded_total = neg_seq_encoded_total[0:pos_seq_encoded_total.shape[0]]
    elif args.data == '80bp':
        args.length = 80
        print("training data type: 80bp containg the repeat - initial position between 1-20")
        
        pos_seq_encoded_chr19 = loadtxt(cwd +  data_dir_path + 'window_80/' + 'chr19.fa_positive_80.csv', delimiter=',')  
        pos_seq_encoded_chr21 = loadtxt(cwd +  data_dir_path + 'window_80/' +'chr21.fa_positive_80.csv', delimiter=',')
        pos_seq_encoded_chr22 = loadtxt(cwd +  data_dir_path + 'window_80/' +'chr22.fa_positive_80.csv', delimiter=',')
        pos_seq_encoded_chr7 = loadtxt(cwd +  data_dir_path + 'window_80/' +'chr7.fa_positive_80.csv', delimiter=',')
        
      #  pos_seq_encoded_chr12_15 = loadtxt(cwd +  data_dir_path + 'window_80/' +'chr12-15.fa_positive_80.csv', delimiter=',')
      #  pos_seq_encoded_chr16_18 = loadtxt(cwd +  data_dir_path + 'window_80/' +'chr16_17_18_20.fa_positive_80.csv', delimiter=',')
        neg_seq_encoded_chr19 = loadtxt(cwd + data_dir_path + 'window_80/' +'chr19.fa_negative_80_uscs.csv', delimiter=',')
        pos_seq_encoded_total = np.concatenate((pos_seq_encoded_chr19, pos_seq_encoded_chr21, pos_seq_encoded_chr22, pos_seq_encoded_chr7), axis = 0)#, pos_seq_encoded_chr16_18, pos_seq_encoded_chr12_15), axis = 0)
        
        neg_seq_encoded_total =  neg_seq_encoded_chr19 #np.append(neg_seq_encoded_chr19, neg_seq_encoded_chr7, axis = 0)
        
    elif args.data == '60bp':
        args.length = 60
        print("training data type: 60bp containg the repeat - initial position between 1-20")
        
     #   pos_seq_encoded_total = loadtxt(cwd + data_dir_path + 'window_60/' +'chr19_21_22_7_8.fa_positive_60.csv', delimiter=',')  
        neg_seq_encoded_total = loadtxt(cwd + data_dir_path + 'window_60/' +'chr19.fa_negative_60_uscs.csv', delimiter=',')
        
        pos_seq_encoded_chr19 = loadtxt(cwd +  data_dir_path + 'window_60_pos_0/' + 'chr19.fa_positive_60_pos0.csv', delimiter=',')  
        pos_seq_encoded_chr21 = loadtxt(cwd +  data_dir_path + 'window_60_pos_0/' +'chr21.fa_positive_60_pos0.csv', delimiter=',')
        pos_seq_encoded_chr22 = loadtxt(cwd +  data_dir_path + 'window_60_pos_0/' +'chr22.fa_positive_60_pos0.csv', delimiter=',')
        pos_seq_encoded_total = np.concatenate((pos_seq_encoded_chr19, pos_seq_encoded_chr21, pos_seq_encoded_chr22), axis = 0)                     
    #    pos_seq_encoded_total = loadtxt(cwd + data_dir_path + 'window_60_pos_0/' + 'chr19_21_22_7.fa_positive_60_pos0.csv', delimiter=',')  
        
    elif args.data == 'sliding':
        args.length = 80
        print("training data type: sliding windows")

        pos_seq_encoded_chr19 = loadtxt(cwd + data_dir_path + 'sliding/' + 'chr19.test_positive_spikes.csv', delimiter=',')  
        pos_seq_encoded_chr19_2 = loadtxt(cwd + data_dir_path + 'sliding/' + 'chr19.test_positive_spikes2.csv', delimiter=',')
        neg_seq_encoded_chr19 = loadtxt(cwd + data_dir_path + 'sliding/' + 'chr19.test_negative_spikes_uscs.csv', delimiter=',')
        neg_seq_encoded_chr19_2 = loadtxt(cwd + data_dir_path + 'sliding/' + 'chr19.test_negative_spikes2_uscs.csv', delimiter=',')
        pos_seq_encoded_total = np.append(pos_seq_encoded_chr19, pos_seq_encoded_chr19_2, axis = 0)
        neg_seq_encoded_total = np.append(neg_seq_encoded_chr19, neg_seq_encoded_chr19_2, axis = 0)
    
    elif args.data == '80bp_st_pos_0_55':
        args.length = 80
        print("training data type: 80bp windows, repeat starts at pos 0-55")

        pos_seq_encoded_total = loadtxt(cwd + data_dir_path + '60bp_stpos_0_55/' + 'chr19_21_22_7.fa_positive_80_stpos0_55.csv', delimiter=',')  
        
        neg_seq_encoded_total = loadtxt(cwd + data_dir_path + 'window_80/' +'chr19.fa_negative_80_uscs.csv', delimiter=',')

    elif args.data == 'full_repeat_80':
        args.length = 80
        print("training data type: 80bp full repeats")#1-20")
        
        pos_seq_encoded_total = loadtxt(cwd + data_dir_path + 'full_repeat_80/' +'chr2_5_6_9.fa_positive_80_full_repeat.csv',
                                        delimiter=',')  

     #   pos_seq_encoded_total = np.concatenate((pos_seq_encoded_1, pos_seq_encoded_2, pos_seq_encoded_3, pos_seq_encoded_4, pos_seq_encoded_5), axis = 0)
        neg_seq_encoded_total = loadtxt(cwd + data_dir_path + 'window_80/' +'chr19.fa_negative_80_uscs.csv', delimiter=',')
    
    
    if args.only_positives == True:
        print("training VAE only on positive samples")
        neg_seq_encoded_train = neg_seq_encoded_total[0:1,:]
    else:
        neg_seq_encoded_train = neg_seq_encoded_total
    
    if args.only_negatives == True:
        print("training VAE only on positive samples")
        pos_seq_encoded_train = pos_seq_encoded_total[0:1,:]
    else:
        pos_seq_encoded_train = pos_seq_encoded_total
        
    print("negative", neg_seq_encoded_train.shape)
    print("positive", pos_seq_encoded_train.shape)
    
    sequence_data = np.concatenate((pos_seq_encoded_train, neg_seq_encoded_train), axis = 0)  
    
    disease = loadtxt(cwd + data_dir_path + 'disease_repeats/' + 'disease_1_45_2420_twos.csv', delimiter=',') 
    
    if args.data == 'before':
        start = 400 - 4*args.length
        end = 400
    else:
        start = 400
        end = 400+4*args.length
        
    disease = disease[:,start:end]
    
    args.test_model = True
    data_disease, args = load_dataset(args, disease)

    for i in range(0,models_no):
        
        hyp = grid[i]
        batch_size = hyp[0]; no_filters_conv1= hyp[1];  filter_size = hyp[2]; stride = hyp[3]; pool_size = hyp[4];
        pool_stride = hyp[5]; lr = hyp[6]; latent_dim = hyp[7]; categorical_dim = hyp[8]; temp = hyp[9]; num_conv = hyp[10];
        filter2_size = hyp[11]; no_filters_conv2 = hyp[12]

        args.batch_size = batch_size
            
        # Train model        
        print('------------------------------')
        print(f'Start Training Model {i}')
        print('------------------------------')
        
        print(batch_size,
        no_filters_conv1,
        filter_size,
        no_filters_conv2, 
        filter2_size,
        stride,
        pool_size,
        pool_stride,
        lr,
        latent_dim,
        categorical_dim,
        temp, num_conv)
        
        if lr == 1e-5:
            epochs = 400
        elif lr == 1e-4:
            epochs = 500
        elif lr == 1e-6:
            epochs = 400
            
        pool_stride = (1, pool_size[1])
            
        args.test_model = False
        train_loader, val_loader, test_loader, args = load_dataset(args, sequence_data, VAE_training=True) #need to load them inside the loop because of the batch_size hyperparameter
        
        print(temp, latent_dim, categorical_dim, no_filters_conv1, no_filters_conv2, filter_size, filter_size, stride, pool_size, pool_stride, args.length, args.act, num_conv)
        
        model = VAE_gumbel_hp_tuning(temp, latent_dim, categorical_dim, no_filters_conv1, no_filters_conv1, filter_size, filter2_size, stride, pool_size, pool_stride, args.length, args.act, num_conv)
        #json_dir = "model0.pt"  
    
   #     model.load_state_dict(torch.load(json_dir, map_location=torch.device('cpu')))    

        if torch.cuda.is_available():
            model.cuda()
            print("Model on GPU")
        
        print("latent dim", latent_dim)
        print(model)
        
        ####### Train VAE #######
        args.test_model = True
        end = min(256*int(pos_seq_encoded_total.shape[0]/256), 256*int(neg_seq_encoded_total.shape[0]/256))

        data_pos, args = load_dataset(args, pos_seq_encoded_total[0:end], VAE_training=False)    
        data_neg, args = load_dataset(args, neg_seq_encoded_total[0:end], VAE_training=False)
        args.test_model = False

        for epoch in range(1, 150 + 1):
            train(epoch, model, train_loader, lr, args)
            test(epoch, model, test_loader, args)
            if epoch%36 == 0:
                train_accuracy_curve, train_f1_curve, val_accuracy_curve, val_f1_curve, obj = classification_latent_space(args, model, data_pos, data_neg)
                args.batch_size = 45
                latent_re = latent_repr(data_disease, model, args)
                latent_re = latent_re.astype(np.float32)
                latent_re = torch.from_numpy(latent_re).cuda() 
               # loss_disease = evaluate(data_disease, model, args) 
               # loss_disease = np.expand_dims(loss_disease, 1)
                
                latent_re = latent_re.cpu().detach().numpy()
                #latent_re = np.concatenate((latent_re, loss_disease), axis=1)
                latent_re = latent_re.astype(np.float32)
                latent_re = torch.from_numpy(latent_re).cuda()
                
                output = obj.model.forward(latent_re) 
                print(latent_re.shape)
                print(output)
                args.batch_size = batch_size
                json_dir = f"model{i}.pt"
                print("Saving data at: " + json_dir)
                torch.save(model.state_dict(), json_dir)
        
        json_dir = f"model{i}.pt"
        print("Saving data at: " + json_dir)
        torch.save(model.state_dict(), json_dir)
        
        ####### Train CNN on the latent space #######
        # classification on the latent space         
        train_accuracy_curve, train_f1_curve, val_accuracy_curve, val_f1_curve, obj = classification_latent_space(args, model, data_pos, data_neg)
        
        args.batch_size = 45
        latent_re = latent_repr(data_disease, model, args)
        latent_re = latent_repr(data_disease, model, args)
        latent_re = latent_re.astype(np.float32)
        latent_re = torch.from_numpy(latent_re).cuda() 
            
        output = obj.model.forward(latent_re) 
        print(output)
        output = output.detach().cpu().numpy()
        
        #Write hyper-parameters and results to csv file
        print('Writing hyper-parameters and results file "{}"'.format(hyp_file_write_path))
        metrics_train = obj.metrics['train']
        metrics_val = obj.metrics['val']
        data = {'Model': i, 'Batch Size': batch_size, 'Num Conv layers': num_conv,
                'Filter no1': no_filters_conv1,'Filter no2': no_filters_conv2, 
                'Conv Filter size0': filter_size[0],  'Conv Filter size': filter_size[1],
                'Conv Filter stride0': stride[0], 'Conv Filter stride': stride[1], 'Pooling size0': pool_size[0],
                'Pooling size': pool_size[1],'Pooling stride0':pool_stride[0], 'Pooling stride':pool_stride[1],
                'LR':lr, 'Latent dim':latent_dim, 'Categorical dim':categorical_dim, ' Temp':temp, 
                'Val Loss': metrics_val['loss'], 'Val Accuracy': metrics_val['acc'], 'Val F1': metrics_val['f1'], 
                'Val Prec': metrics_val['prec'], 'Val Recall': metrics_val['recall'], 'Train Loss': metrics_train['loss'],
                'Train Accuracy': metrics_train['acc'], 'Train F1': metrics_train['f1'], 'Train Prec': metrics_train['prec'], 
                'Train Recall': metrics_train['recall'], 'D1': output[0], 'D2': output[1], 'D3': output[2], 'D4': output[3], 
                'D5': output[4], 'D6': output[5], 'D7': output[6], 'D8': output[7], 'D9': output[8], 'D10': output[9], 'D11': output[10], 
               'D12': output[11], 'D13': output[12], 'D14': output[13], 'D15': output[14], 'D16': output[15], 'D17': output[16], 
                'D18': output[17], 'D19': output[18], 'D20': output[19], 'D21': output[20], 'D22': output[21],
               'D23': output[22], 'D24': output[23], 'D25': output[24], 'D26': output[25], 'D27': output[26], 
                'D28': output[27], 'D29': output[28], 'D30': output[29], 'D31': output[30], 'D32': output[31],
               'D33': output[32], 'D34': output[33], 'D35': output[34], 'D36': output[35], 'D37': output[36], 
                'D38': output[37], 'D39': output[38], 'D40': output[39], 'D41': output[40], 'D42': output[41],
               'D43': output[42], 'D44': output[43],'D5': output[44]}
                
        pd.DataFrame(data, index=[0]).to_csv(hyp_file_write_path, mode='a', header=True)
