import torch
import argparse
import time
import torch
import torch.utils.data
import torch.optim as optim
import numpy as np
import math
import random
import torch.utils.data as data_utils

from load_data import load_dataset
import json
import pathlib
import time
from numpy import loadtxt, savetxt
import matplotlib.pyplot as plt

import numpy as np 
import pandas as pd
from pandas.plotting import scatter_matrix
import seaborn as sns
from vae_gumbel_softmax import VAE_gumbel, VAE_gumbel_hp_tuning, latent_repr

parser = argparse.ArgumentParser(description='PyTorch Normalizing flows')

parser.add_argument('-d', '--dataset', type=str, default='mnist', choices=['mnist'],
                    metavar='DATASET',
                    help='Dataset choice.')

parser.add_argument('-nc', '--no_cuda', action='store_true', default=False,
                    help='disables CUDA training')

parser.add_argument('--manual_seed', type=int, help='manual seed, if not given resorts to random seed.')

parser.add_argument('-li', '--log_interval', type=int, default=10, metavar='LOG_INTERVAL',
                    help='how many batches to wait before logging training status')

parser.add_argument('-od', '--out_dir', type=str, default='logs/', metavar='OUT_DIR',
                    help='output directory for model snapshots etc.')

fp = parser.add_mutually_exclusive_group(required=False)
fp.add_argument('-te', '--testing', action='store_true', dest='testing',
                help='evaluate on test set after training')
fp.add_argument('-va', '--validation', action='store_false', dest='testing',
                help='only evaluate on validation set')
parser.set_defaults(testing=False)

# optimization settings
parser.add_argument('-e', '--epochs', type=int, default=1, metavar='EPOCHS',
                    help='number of epochs to train (default: 1000)')
parser.add_argument('-es', '--early_stopping_epochs', type=int, default=100, metavar='EARLY_STOPPING',
                    help='number of early stopping epochs')

parser.add_argument('-bs', '--batch_size', type=int, default=1, metavar='BATCH_SIZE',
                    help='input batch size for training (default: 100)')
parser.add_argument('-lr', '--learning_rate', type=float, default=0.00001, metavar='LEARNING_RATE',
                    help='learning rate')

parser.add_argument('-a', '--anneal', type=str, default="kl", choices= ["std", "off", "kl"], help="beta annealing scheme")
parser.add_argument('--max_beta', type=float, default=1., metavar='MB',
                    help='max beta for warm-up')
parser.add_argument('--min_beta', type=float, default=0.0, metavar='MB',
                    help='min beta for warm-up')
parser.add_argument('-f', '--flow', type=str, default='planar', choices=['planar', 'NICE', 'NICE_MLP', 'syl_orthogonal', 'real' ])
parser.add_argument('-nf', '--num_flows', type=int, default=0,
                    metavar='NUM_FLOWS', help='Number of flow layers, ignored in absence of flows')
parser.add_argument('-nov', '--num_orthonormal_vec', type=int, default=8, metavar='NUM_ORTHONORMAL_VEC',
                    help='For orthogonal flow: number of orthogonal vectors (M)')
parser.add_argument('--syl_ortho_invertible', type=bool, default=False, metavar='ORTHO_FLOW_INVERT',
                    help='select if we need the sylvester orthogonal flow to be invertible')
parser.add_argument('--z_size', type=int, default=40, metavar='ZSIZE',
                    help='how many stochastic hidden units')
parser.add_argument('--flat_seq_size', type=int, default=480, metavar='FSSIZE',
                    help='size of flattened sequence')
parser.add_argument('--encoder_dim', type=int, default=40, metavar='ESIZE',
                    help='output feature dim of encoder')
parser.add_argument('--decoder_dim', type=int, default=480, metavar='DSIZE',
                    help='output feature dim of decoder') 

parser.add_argument('--cnn_enc', type=bool, default=True, metavar='cnn_enc',
                    help='encoder cnn')

parser.add_argument('--cnn_dec', type=bool, default=True, metavar='cnn_dec',
                    help='decoder cnn')
parser.add_argument('--test_model', type=bool, default=True, metavar='test_model',
                    help='test model')
# gpu/cpu
parser.add_argument('--gpu_num', type=int, default=0, metavar='GPU', help='choose GPU to run on.')

parser.add_argument('--temp', type=float, default=1.0, metavar='S',
                    help='tau(temperature) (default: 1.0)')

args = parser.parse_args()
args.cuda = not args.no_cuda and torch.cuda.is_available()

if args.manual_seed is None:
    args.manual_seed = random.randint(1, 100000)
random.seed(args.manual_seed)
torch.manual_seed(args.manual_seed)
np.random.seed(args.manual_seed)

if args.cuda:
    torch.cuda.set_device(args.gpu_num)

    
def run(args):

    sns.color_palette("hls", 8)

#    pos_seq_encoded_chr19 = loadtxt('chr19.test_positive_spikes.csv', delimiter=',')
   # pos_seq_encoded_chr19 = loadtxt('chr19.fa_positive_80.csv', delimiter=',')
    pos_seq_encoded_chr19 = loadtxt('chr19_21_22_7_8.fa_positive_60.csv', delimiter=',')  

    N = pos_seq_encoded_chr19.shape[0]
    indices = np.arange(N)
    np.random.shuffle(indices)
    pos_seq_encoded_chr19 = pos_seq_encoded_chr19[indices, :]
    data1, args = load_dataset(args, pos_seq_encoded_chr19[0:10000])
    
    print(pos_seq_encoded_chr19.shape)
    
    
    #neg_seq_encoded_chr19 = loadtxt('chr19.test_negative_spikes_uscs.csv', delimiter=',')
#    neg_seq_encoded_chr19 = loadtxt('chr19.fa_negative_80_uscs.csv', delimiter=',')
    neg_seq_encoded_chr19 = loadtxt('chr19.fa_negative_60_uscs.csv', delimiter=',')  

    N = neg_seq_encoded_chr19.shape[0]
    indices = np.arange(N)
    np.random.shuffle(indices)
    neg_seq_encoded_chr19 = neg_seq_encoded_chr19[indices, :]
    
 #   neg_seq_encoded_chr19 = loadtxt('chr19.test_negative_spikes.csv', delimiter=',')
    data2, args = load_dataset(args, neg_seq_encoded_chr19[0:10000])
    
    disease1 = loadtxt('disease_loci_1.csv', delimiter=',')
    
    disease2 = loadtxt('disease_loci_2.csv', delimiter=',')
    disease3 = loadtxt('disease_loci_3.csv', delimiter=',')
    disease4 = loadtxt('disease_loci_4.csv', delimiter=',')
    disease5 = loadtxt('disease_loci_5.csv', delimiter=',')
    disease6 = loadtxt('disease_loci_6.csv', delimiter=',')
    disease7 = loadtxt('disease_loci_7.csv', delimiter=',')
    disease8 = loadtxt('disease_loci_8.csv', delimiter=',')
    disease9 = loadtxt('disease_loci_9.csv', delimiter=',')
    
    
    
    disease1 = np.expand_dims(disease1[400:640], 0)
 #   disease2 = np.expand_dims(disease[100:420], 0)
 #   disease3 = np.expand_dims(disease[200:520], 0)
 #   disease4 = np.expand_dims(disease[300:620], 0)
 #   disease5 = np.expand_dims(disease[400:720], 0)
    
    disease2 = np.expand_dims(disease2[400:640], 0)
    disease3 = np.expand_dims(disease3[400:640], 0)
    disease4 = np.expand_dims(disease4[400:640], 0)
    disease5 = np.expand_dims(disease5[400:640], 0)
    disease6 = np.expand_dims(disease6[400:640], 0)
    disease7 = np.expand_dims(disease7[400:640], 0)
    disease8 = np.expand_dims(disease8[400:640], 0)
    
    disease9 = np.expand_dims(disease9[400:640], 0)
    
    disease = np.concatenate((disease1, disease2, disease3, disease4, disease5, disease6, disease7, disease8, disease9), axis = 0)#, disease6, disease7, disease8, disease9), axis = 0)
    
    print(disease.shape)
    
    data3, args = load_dataset(args, disease)
#    data4, args = load_dataset(args, disease2)
#    data5, args = load_dataset(args, disease3)
#    data6, args = load_dataset(args, disease4)
#    data7, args = load_dataset(args, disease5)
   # model = VAE_gumbel(args.temp, 4, 2)
    model = VAE_gumbel_hp_tuning(0.7, 4, 2, 25, 25, 3, 1, 3, 2)

    
    json_dir = "model0.pt"
    model.load_state_dict(torch.load(json_dir))
    if torch.cuda.is_available():
        model.cuda()
        print("Model on GPU")
        
    print(model)
 #   la = evaluate(data1, model, args)
    latent1 = latent_repr(data1, model, args)
    latent2 = latent_repr(data2, model, args)
    latent3 = latent_repr(data3, model, args)
  #  latent4 = latent_repr(data4, model, args)
  #  latent5 = latent_repr(data5, model, args)
  #  latent6 = latent_repr(data6, model, args)
  #  latent7 = latent_repr(data7, model, args)
    
 #   print("shape before ", latent1.shape)
    x_1 = np.ones((latent1.shape[0], 1))
    x_2 = np.ones((latent2.shape[0], 1))
    x_3 = np.ones((latent3.shape[0], 1))
   # x_4 = np.ones((latent4.shape[0], 1))
   # x_5 = np.ones((latent5.shape[0], 1))
   # x_6 = np.ones((latent6.shape[0], 1))
   # x_7 = np.ones((latent7.shape[0], 1))
  #  print(latent1)
    data_TSNE = np.concatenate((latent1, latent2, latent3), axis=0)#, latent4, latent5, latent6, latent7), axis=0)
    ys = np.concatenate((1*x_1, 2*x_2, 3*x_3), axis=0) #, 4*x_4, 5*x_5, 6*x_6, 7*x_7), axis=0)
    
    
    latent1 = np.concatenate((latent1, 1*x_1), axis=1)
  #  print("shape after ", latent1.shape)
    latent2 = np.concatenate((latent2, 2*x_2), axis=1)
    latent3 = np.concatenate((latent3, 3*x_3), axis=1)
  #  latent4 = np.concatenate((latent4, 4*x_3), axis=1)
  #  latent5 = np.concatenate((latent5, 5*x_3), axis=1)
  #  latent6 = np.concatenate((latent6, 6*x_3), axis=1)
  #  latent7 = np.concatenate((latent7, 7*x_3), axis=1)
    
    data = np.concatenate(( latent1, latent2, latent3), axis=0) #, latent4, latent5, latent6, latent7), axis=0)
    df1 = pd.DataFrame(data, columns=['1','2','3','4','5','6','7'])#,'8','9', '10', '11', '12', '13', '14', '15', '16', '17'])#

 #   print(latent1.shape)
 #   print(latent1[:,0])
    #plot 2 dim of latent space
    
    sns.pairplot(df1, hue='7')
    #pd.plotting.scatter_matrix(df1)
    #pd.plotting.scatter_matrix(df2)
    #pd.plotting.scatter_matrix(df3)
    
    # run TSNE when number of latent dims exceeds 2
    from sklearn.manifold import TSNE
    zs2 = TSNE().fit_transform(data_TSNE)
    zs2_mean = zs2.mean(0)
    zs2_std = zs2.std(0)


    # display a 2D plot of the classes in the latent space
    fig = plt.figure(figsize=(6,6))
    ax = plt.gca()

    colors = []
    for k in range(1, 4):
        m = (ys == k)
        print(m.shape)
        m = np.squeeze(m,1)
        p = ax.scatter(zs2[m, 0], zs2[m, 1], label='y=%d' % k, alpha=0.5, s=5)
        colors.append(p.get_facecolor())
    ax.legend()

    fig.tight_layout()
    #fig.savefig('images/%s-%02d-encodings.png' % (MODEL_NAME, NUM_STYLE), dpi=300)

    x_lim = ax.get_xlim()
    y_lim = ax.get_ylim()
      #  plt.scatter(latent1[:,0], latent1[:,1])
      #  plt.scatter(latent2[:,0], latent2[:,1])
      #  plt.scatter(latent3[:,0], latent3[:,1])
      #  plt.show()
if __name__ == "__main__":
    

    run(args)